/*
 * $Id: ppcPixmapFS.c,v 1.2 1991/10/05 01:25:32 mtranle Exp $
 *
 * Copyright IBM Corporation 1987,1988,1989
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that 
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of IBM not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * IBM DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * IBM BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
*/

/******************************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#include "X.h"
#include "misc.h"
#include "gcstruct.h"
#include "window.h"
#include "pixmapstr.h"
#include "scrnintstr.h"
#include "windowstr.h"

#include "mfb.h"
#include "maskbits.h"

#include "OScompiler.h"

#include "ppc.h"
#include "ppcSpMcro.h"

#include "ibmTrace.h"
extern int mfbGCPrivateIndex;

#define LeftMostBitInScreenLongWord SCRLEFT( 0xFFFFFFFF, 31 )

/* Binary search to figure out what to do for the raster op.  It may
 * do 5 comparisons, but at least it does no function calls 
 * Special cases copy because it's so frequent 
 * XXX - can't use this in many cases because it has no plane mask.
 */
#undef  DoRop
#define DoRop(alu, src, dst) \
( ((alu) == GXcopy) ? (src) : \
    (((alu) >= GXnor) ? \
     (((alu) >= GXcopyInverted) ? \
       (((alu) >= GXnand) ? \
         (((alu) == GXnand) ? ~((src) & (dst)) : ~0) : \
         (((alu) == GXcopyInverted) ? ~(src) : (~(src) | (dst)))) : \
       (((alu) >= GXinvert) ? \
	 (((alu) == GXinvert) ? ~(dst) : ((src) | ~(dst))) : \
	 (((alu) == GXnor) ? ~((src) | (dst)) : (~(src) ^ (dst)))) ) : \
     (((alu) >= GXandInverted) ? \
       (((alu) >= GXxor) ? \
	 (((alu) == GXxor) ? ((src) ^ (dst)) : ((src) | (dst))) : \
	 (((alu) == GXnoop) ? (dst) : (~(src) & (dst)))) : \
       (((alu) >= GXandReverse) ? \
	 (((alu) == GXandReverse) ? ((src) & ~(dst)) : (src)) : \
	 (((alu) == GXand) ? ((src) & (dst)) : 0)))  ) )

/*
********** ********** ********** ********** ********** ********** **********
   these routines all clip.  they assume that anything that has called
them has already translated the points (i.e. pGC->miTranslate is
non-zero, which is howit gets set in mfbCreateGC().)

   the number of new scnalines created by clipping ==
MaxRectsPerBand * nSpans.
********** ********** ********** ********** ********** ********** **********
*/

void
ppcSolidPixmapFS( pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted )
     DrawablePtr pDrawable ;
     GCPtr	pGC ;
     int	nInit ;			/* number of spans to fill */
     DDXPointPtr pptInit ;		/* pointer to list of start points */
     int	*pwidthInit ;		/* pointer to list of n widths */
     int 	fSorted ;
{
    unsigned long int pm, npm ;
    unsigned long int fg ;
    int 	alu ;
				/* next three parameters are post-clip */
    int 	n ;		/* number of spans to fill */
    DDXPointPtr ppt ;		/* pointer to list of start points */
    int 	*pwidth ;	/* pointer to list of n widths */
    unsigned char *addrl ;	/* pointer to current longword in bitmap */
    int 	i ;
    int 	*pwidthFree ;	/* copies of the pointers to free */
    DDXPointPtr pptFree ;
    ppcPrivGC   *priv = (ppcPrivGC *)pGC->devPrivates[mfbGCPrivateIndex].ptr;

    if ( pDrawable->type != DRAWABLE_PIXMAP )
    {
	ErrorF("ppcSolidPixmapFS: drawable is not a pixmap\n") ;
	return ;
    }

    if( (alu = priv->colorRrop.alu) == GXnoop )
	return ;

    n = nInit *	miFindMaxBand( priv->pCompositeClip) ;
    if( !(pwidth = (int *) ALLOCATE_LOCAL(n * sizeof(int))) )
	return ;
    pwidthFree = pwidth ;

    if( !( ppt = (DDXPointRec *) ALLOCATE_LOCAL( n * sizeof( DDXPointRec ))) )
    {
	DEALLOCATE_LOCAL( pwidth ) ;
	return ;
    }
    pptFree = ppt ;

    n = miClipSpans( priv->pCompositeClip, pptInit, pwidthInit, nInit, ppt,
		     pwidth, fSorted ) ;

    pm = priv->colorRrop.planemask ;
    fg = priv->colorRrop.fgPixel ;
    npm = ( ~ pm ) & ( ( 1 << pDrawable->depth ) - 1 ) ;

    for ( ; n-- ; ppt++, pwidth++ )
    {
        addrl = (( (unsigned char *) ((PixmapPtr) pDrawable )->devPrivate.ptr)
		 + (ppt->y * ( (int) ((PixmapPtr) pDrawable )->devKind ) )
		 + ppt->x ) ;
	for ( i = *pwidth ; i-- ; addrl++ )
	    *addrl = ( *addrl & npm ) | ( pm & DoRop( alu, fg, *addrl ) ) ;
    }
    DEALLOCATE_LOCAL( pptFree ) ;
    DEALLOCATE_LOCAL( pwidthFree ) ;
    return ;
}

void
ppcStipplePixmapFS( pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted )
     DrawablePtr pDrawable ;
     GC *pGC ;
     int nInit ;		/* number of spans to fill */
     DDXPointPtr pptInit ;	/* pointer to list of start points */
     int *pwidthInit ;		/* pointer to list of n widths */
     int fSorted ;
{
    unsigned char *pdst ;	/* pointer to current word in bitmap */
    int *psrc ;			/* pointer to current word in tile */
    unsigned long int pm, npm ;
    unsigned long int fg ;
    int alu ;
				/* next three parameters are post-clip */
    int n ;			/* number of spans to fill */
    DDXPointPtr ppt ;		/* pointer to list of start points */
    int *pwidth ;		/* pointer to list of n widths */
    PixmapPtr	pTile ;		/* pointer to tile we want to fill with */
    int		width, xSrc, ySrc ;
    int 	tlwidth, tileWidth, *psrcT ;
    int *pwidthFree ;		/* copies of the pointers to free */
    DDXPointPtr pptFree ;
    int xoff, count, stip, i ;
    ppcPrivGC   *priv = (ppcPrivGC *)pGC->devPrivates[mfbGCPrivateIndex].ptr;

    if ( pDrawable->type != DRAWABLE_PIXMAP )
    {
	ErrorF( "ppcStippleWindowFS: drawable is not a pixmap\n") ;
	return ;
    }
    if ( pGC->stipple->drawable.depth != 1 )
    {
	ErrorF( "ppcStippleFS: bad depth\ntype = %d, depth = %d\n",
	       pDrawable->type, pGC->stipple->drawable.depth ) ;
	return ;
    }

    if ( ( alu = priv->colorRrop.alu ) == GXnoop )
	return ;

    SETSPANPTRS( nInit, n, pwidthInit, pwidth, pptInit,
		ppt, pwidthFree, pptFree, fSorted ) ;

    pm        = priv->colorRrop.planemask ;
    fg        = priv->colorRrop.fgPixel ;
    npm       = ( ~ pm ) & ( ( 1 << pDrawable->depth ) - 1 ) ;
    pTile     = pGC->stipple ;
    tlwidth   = pTile->devKind >> 2 ;
    tileWidth = pTile->drawable.width ;

    /* this replaces rotating the stipple.  Instead, we just adjust the offset
     * at which we start grabbing bits from the stipple */
    xSrc = (pGC->patOrg.x % pTile->drawable.width)  - pTile->drawable.width ;
    ySrc = (pGC->patOrg.y % pTile->drawable.height) - pTile->drawable.height ;

    for ( ; n-- ; ppt++, pwidth++ )
    {
        pdst = (( (unsigned char *) ((PixmapPtr) pDrawable )->devPrivate.ptr )
		+ ( ppt->y * ((int) ((PixmapPtr) pDrawable )->devKind ) ) 
		+ ppt->x ) ;
        psrcT = ((int *) pTile->devPrivate.ptr
		 + ( ( ( ppt->y - ySrc ) % pTile->drawable.height ) *
		     tlwidth ) ) ;

	xoff = ( ppt->x - xSrc ) % tileWidth ;
	psrc = psrcT + ( xoff >> 5 ) ;
	for ( width = *pwidth ; width ; psrc++, width -= count )
	{
	    if ( psrc >= ( psrcT + tlwidth ) )
		psrc = psrcT ;
	    if ( ( psrc != ( psrcT + tlwidth - 1 ) ) ||
		!( count = ( tileWidth & 0x1F ) ) )
		count = 32 ;
	    if ( count > width )
		count = width ;
	    getbits( psrc, xoff & 0x1F, count, stip ) ;
	    for ( i = count ; i-- ; pdst++, stip = SCRLEFT( stip, 1 ) )
		if ( stip & LeftMostBitInScreenLongWord )
		    *pdst = (*pdst & npm) | (pm & DoRop( alu, fg, *pdst)) ;
	}
    }
    DEALLOCATE_LOCAL( pptFree ) ;
    DEALLOCATE_LOCAL( pwidthFree ) ;
    return ;
}

void
ppcOpStipplePixmapFS( pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted )
     DrawablePtr pDrawable ;
     GC *pGC ;
     int nInit ;		/* number of spans to fill */
     DDXPointPtr pptInit ;	/* pointer to list of start points */
     int *pwidthInit ;		/* pointer to list of n widths */
     int fSorted ;
{
    unsigned char *pdst ;	/* pointer to current word in bitmap */
    unsigned long int pm, npm ;
    unsigned long int fg, bg ;
    int 	alu ;
				/* next three parameters are post-clip */
    int 	n ;		/* number of spans to fill */
    DDXPointPtr ppt ;		/* pointer to list of start points */
    int 	*pwidth ;	/* pointer to list of n widths */
    int 	*psrc ;		/* pointer to current word in tile */
    PixmapPtr 	pTile ;		/* pointer to tile we want to fill with */
    int		width, xSrc, ySrc ;
    int 	tlwidth, tileWidth, *psrcT ;
    int 	*pwidthFree ;	/* copies of the pointers to free */
    DDXPointPtr pptFree ;
    int 	xoff, count, stip, i ;
    ppcPrivGC   *priv = (ppcPrivGC *)pGC->devPrivates[mfbGCPrivateIndex].ptr;

    if ( pGC->stipple->drawable.depth != 1 )
    {
	ErrorF( "ppcOpStipplePixmapFS: bad depth\ntype = %d, depth = %d\n",
	       pDrawable->type, pGC->stipple->drawable.depth ) ;
	return ;
    }

    if ( ( alu = priv->colorRrop.alu ) == GXnoop )
	return ;

    SETSPANPTRS( nInit, n, pwidthInit, pwidth, pptInit,
		ppt, pwidthFree, pptFree, fSorted ) ;

    fg        = priv->colorRrop.fgPixel ;
    bg        = priv->colorRrop.bgPixel ;
    pm        = priv->colorRrop.planemask ;
    npm       = ( ~ pm ) & ( ( 1 << pDrawable->depth ) - 1 ) ;
    pTile     = pGC->stipple ;
    tlwidth   = pTile->devKind >> 2 ;
    tileWidth = pTile->drawable.width ;


    /* this replaces rotating the stipple.  Instead, we just adjust the offset
     * at which we start grabbing bits from the stipple */
    xSrc = (pGC->patOrg.x % pTile->drawable.width)  - pTile->drawable.width ;
    ySrc = (pGC->patOrg.y % pTile->drawable.height) - pTile->drawable.height ;

    for ( ; n-- ; ppt++, pwidth++ )
    {
        pdst = (( (unsigned char *) ((PixmapPtr) pDrawable )->devPrivate.ptr )
		+ (ppt->y * ( (int) ((PixmapPtr) pDrawable )->devKind ) )
		+ ppt->x ) ;
        psrcT = ((int *) pTile->devPrivate.ptr
		 + ( ( ( ppt->y - ySrc ) % pTile->drawable.height ) *
		     tlwidth )) ;

	xoff = ( ppt->x - xSrc ) % tileWidth ;
	psrc = psrcT + ( xoff >> 5 ) ;
	for ( width = *pwidth ; width ; psrc++, width -= count )
	{
	    if ( psrc >= ( psrcT + tlwidth ) )
		psrc = psrcT ;
	    if ( ( psrc != ( psrcT + tlwidth - 1 ) ) ||
		!( count = tileWidth & 0x1F ) )
		count = 32 ;
	    if ( count > width )
		count = width ;
	    getbits( psrc, xoff & 0x1F, count, stip ) ;
	    for ( i = count ; i-- ; pdst++, stip = SCRLEFT( stip, 1 ) )
		if ( stip & LeftMostBitInScreenLongWord )
		    *pdst = (*pdst & npm) | (pm & DoRop( alu, fg, *pdst)) ;
		else
		    *pdst = (*pdst & npm) | (pm & DoRop( alu, bg, *pdst)) ;
	}
    }
    DEALLOCATE_LOCAL( pptFree ) ;
    DEALLOCATE_LOCAL( pwidthFree ) ;
    return ;
}

void
ppcTilePixmapFS( pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted )
     DrawablePtr pDrawable ;
     GC *pGC ;
     int nInit ;		/* number of spans to fill */
     DDXPointPtr pptInit ;	/* pointer to list of start points */
     int *pwidthInit ;		/* pointer to list of n widths */
     int fSorted ;
{
    DDXPointPtr ppt ;		/* pointer to list of start points */
    int 	*pwidth ;	/* pointer to list of n widths */
    unsigned char *pdst ;	/* pointer to current word in bitmap */
    unsigned char *psrc ;	/* pointer to current word in tile */
    PixmapPtr 	pTile ;		/* pointer to tile we want to fill with */
    int 	i, xSrc, ySrc ;
    int 	alu ;
    unsigned char pm, npm ;
				/* next three parameters are post-clip */
    int 	n ;		/* number of spans to fill */
    int 	tileWidth ;
    unsigned char *psrcT ;
    int 	*pwidthFree ;	/* copies of the pointers to free */
    DDXPointPtr pptFree ;
    ppcPrivGC 	*priv = (ppcPrivGC *)pGC->devPrivates[mfbGCPrivateIndex].ptr;

    if ( (pDrawable->depth == 1 ) && ( pDrawable->type == DRAWABLE_PIXMAP) )
    {
	mfbTileFS( pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted ) ;
	return ;
    }
    if ( !ppcDepthOK( pDrawable, pGC->tile.pixmap->drawable.depth ) )
    {
	ErrorF( "ppcTileFS: bad depth\ntype = %d, depth = %d\n",
	       pDrawable->type, pDrawable->depth) ;
	return ;
    }

    if ( ( alu = priv->colorRrop.alu ) == GXnoop )
	return ;

    SETSPANPTRS( nInit, n, pwidthInit, pwidth, pptInit,
		ppt, pwidthFree, pptFree, fSorted ) ;

    /* the following code is for 8 bits per pixel addressable memory only */
    pm        = priv->colorRrop.planemask ;
    npm       = ( ~ pm ) & ( ( 1 << pDrawable->depth ) - 1 ) ;
    pTile     = pGC->tile.pixmap ;
    tileWidth = pTile->drawable.width ;

    /* this replaces rotating the tile. Instead we just adjust the offset
     * at which we start grabbing bits from the tile */
    xSrc = (pGC->patOrg.x % pTile->drawable.width) - pTile->drawable.width ;
    ySrc = (pGC->patOrg.y % pTile->drawable.height) - pTile->drawable.height ;

    for ( ; n-- ; ppt++, pwidth++ )
    {
        pdst = (( (unsigned char *) ( (PixmapPtr) pDrawable )->devPrivate.ptr )
		+ ( ppt->y * ( (int) ( (PixmapPtr) pDrawable )->devKind ) )
		+ ppt->x ) ;
        psrcT = ((unsigned char *) pTile->devPrivate.ptr
		 + ( ( ( ppt->y - ySrc ) % pTile->drawable.height ) *
		    pTile->devKind ) ) ;

	psrc = psrcT + ( ( ppt->x - xSrc ) % tileWidth ) ;
	for ( i = *pwidth ; i-- ; pdst++, psrc++ )
	{
	    if ( psrc >= ( psrcT + tileWidth ) )
		psrc = psrcT ;
	    *pdst = ( *pdst & npm ) | ( pm & DoRop( alu, *psrc, *pdst ) ) ;
	}
    }
    DEALLOCATE_LOCAL( pptFree ) ;
    DEALLOCATE_LOCAL( pwidthFree ) ;
}
